#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys      # platform
import time     # sleep
import serial   # serial
import datetime

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

#---------------------------------------------------------------------------
# 共通処理：ログ出力
#---------------------------------------------------------------------------
def putLog(msg):
    if verbose == 1:
        print('%s ' % datetime.datetime.now(), end='')
        print(msg)

#---------------------------------------------------------------------------
# 共通処理：デバッグ用ダンプ出力
#---------------------------------------------------------------------------
def debugDump(bytes, addr = 0):
    if verbose == 1:
        dump(bytes, addr)
        print('')

#---------------------------------------------------------------------------
# 共通処理：byte配列をダンプする
#---------------------------------------------------------------------------
def dump(bytes, addr = 0):
    memory_address = 0 + addr
    ascii_string = ""
    print("Offset   00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F ASCII")
    print("-------- ----------------------------------------------- ----------------")
    
    for byte in bytes:
        ascii_string = ascii_string + chr(validate_byte_as_printable(byte))
        if memory_address%16 == 0:
            print(format(memory_address, '08X'),end = '')
            print(" " + hex(byte)[2:].zfill(2), end='')
        elif memory_address%16 == 15:
            print(" " + hex(byte)[2:].zfill(2),end='')
            print(" " + ascii_string)
            ascii_string = ""
        else:
            print(" " + hex(byte)[2:].zfill(2), end='')
        memory_address = memory_address + 1

def validate_byte_as_printable(byte):
    ## ascii 文字列化をチェック. asciiでなければ '.' を返す ##
    if is_character_printable(byte):
        return byte
    else:
        return 46

def is_character_printable(s):
    ## asciiの英数字・記号の文字列かどうかを判別し、真偽値を返します。
    #if s < 126 and s >= 33:
    if s < 126 and s >= 0x20:
        return True 

#---------------------------------------------------------------------------
# シリアルポート列挙ルーチン
#---------------------------------------------------------------------------
def serial_ports():
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(256)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result


#---------------------------------------------------------------------------
# 値を入力させる
#   不正な値の場合、繰り返し
#   戻り値：入力した値
#---------------------------------------------------------------------------
def inputValue(message):
    while True:
        print(message, end='')
        stAddrStr = input()
        if len(stAddrStr) == 0:
            continue
        else:
            stAddr = int(stAddrStr)
            break
    return stAddr

#---------------------------------------------------------------------------
# 16進の値を入力させる
#   不正な値の場合、繰り返し
#   戻り値：入力した値
#---------------------------------------------------------------------------
def inputHexValue(message):
    while True:
        print(message, end='')
        stAddrStr = input()
        if len(stAddrStr) == 0:
            continue
        else:
            if len(stAddrStr) > 2 and stAddrStr.startswith('0x'):
                stAddr = int(stAddrStr, 0)
            else:
                stAddr = int(stAddrStr, 16)
        break
    return stAddr

#---------------------------------------------------------------------------
# ファイル名をユーザに入力させる
#   不正な値(長さ0)の場合、繰り返し
#   戻り値：入力した値
#
#   ※今後、存在チェックを入れた方がいいかも？
#---------------------------------------------------------------------------
def inputFilename(message):
    while True:
        print(message, end='')
        fileNameStr = input()
        if len(fileNameStr) == 0:
            continue
        break

    return fileNameStr

#---------------------------------------------------------------------------
# 指定アドレスから指定サイズデータ読み取り
#       戻り値：2つの値のタプル
#               retCode: 
#                    0:成功
#                   -1:読み取り要求コマンドの応答タイムアウト
#                   -2:読み取り要求コマンドに肯定応答でない応答
#                   -3:読み取りデータ本体受信タイムアウト
#                   -4:ラストの状態要求コマンド応答が不正
#               retAry:  成功の時、読み取ったデータ
#
#       メモ：
#---------------------------------------------------------------------------
def readSize5(s, addrInt, sizeInt):
    retCode = 0
    retAry  = []
    
    addr = addrInt.to_bytes(4, 'little')
    size = sizeInt.to_bytes(4, 'little')

    # 入出力バッファクリア
    if s.in_waiting > 0:
        putLog('未処理の受信データあり: %d' % s.in_waiting)
        readStr = s.read(s.in_waiting)
        readBytes = bytes(readStr)
        putLog('未処理データ受信。')
        debugDump(readBytes)

    else:
        putLog('未受信データ無し。')

    if s.out_waiting > 0:
        putLog('送信バッファにデータあり: %d' % s.out_waiting)
        s.reset_output_buffer()
    else:
        putLog('未送信データ無し。')


    # 1. 読み取り要求コマンドに肯定応答が来るまで待つ処理
    waitCount = 0
    while True:
        # 1-1. 読み取り要求コマンド送信
        putLog('読み取り要求コマンド送信。')
        s.write(bytes([
            0x0D, 0x00, 0x00, 0x00,             # 読み込み
            0x08, 0x00, 0x00, 0x00,             # 不明 8
            size[0], size[1], size[2], size[3], # サイズ？
            addr[0], addr[1], addr[2], addr[3], # 開始アドレス
            size[0], size[1], size[2], size[3], # サイズ？
            ]))

        # 1-2. 読み取り要求コマンド応答待ち
        respWaitCount = 0
        while True:
            if s.in_waiting > 0:
                putLog('受信データあり: %d' % s.in_waiting)
                break
            else:
                respWaitCount = respWaitCount + 1
                putLog('0.01秒待機。')
                time.sleep(0.01)
                if(respWaitCount > 40):
                    print('ERR: 読み取り要求コマンド応答タイムアウト')
                    return -1, retAry

        # 1-3. 読み取り要求コマンド応答の受信と判定
#        readStr = s.read(s.in_waiting)
        readStr = s.read(8)
        readBytes = bytes(readStr)
        putLog('読み取り要求コマンド応答受信。')
        debugDump(readBytes)

        if readBytes == bytes([0,0,0,0,size[0],size[1],size[2],size[3]]):
            putLog('肯定応答を受信。')
            break
        else:
            putLog('肯定応答でない応答を受信。')
            print('ERR: 読み取り要求コマンドに肯定応答でない応答')
            return -2, retAry

    # 2. 状態要求コマンド送信
    # 実機と異なり、ここで、いつまで待ってもデータ本体が来ない。
    # 催促のため(！？)状態要求コマンドを送信する

    # 状態要求コマンド送信
    putLog('読み取りデータ未着のため状態要求コマンド送信。')
    s.write(bytes([
        0x06, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,]))
    # 応答待ち
    waitCount = 0
    while True:
        if s.in_waiting > 0:
            putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            waitCount = waitCount + 1
            if(waitCount > 40):
                retCode = -3
                print('ERR: 読み取りデータ本体受信タイムアウト')
                return retCode, retAry
            putLog('0.01秒待機。')
            time.sleep(0.01)

    # 3. 読み取りデータ本体受信
    #   ただし、前項で状態要求を出していた場合、
    #   読み取りデータに続いて、状態応答が来るので
    #   忘れずに受信する。忘れるとシーケンスが破綻する
    readStr = s.read(sizeInt)
    readBytes = bytes(readStr)
    putLog('読み取りデータ本体受信。')
    debugDump(readBytes, addrInt)
    retAry.extend(readBytes)

    # 4. 状態要求コマンド応答受信
    readStr = s.read(12)
    readBytes = bytes(readStr)
    putLog('状態要求コマンド応答受信。')
    debugDump(readBytes)

    if readBytes == bytes([0xff,0xff,0xff,0xff,0,0,0,0]):
        putLog('ERR: 状態要求コマンド応答が不正。')
        return -4, retAry

    return retCode, retAry


#---------------------------------------------------------------------------
# 指定アドレスから指定サイズデータ読み取り
#       戻り値：2つの値のタプル
#               retCode: 
#                    0:成功
#                   -1:読み取り要求コマンドの応答タイムアウト
#                   -2:読み取り要求コマンドに肯定応答でない応答
#                   -3:読み取りデータ本体受信タイムアウト
#                   -4:ラストの状態要求コマンド応答が不正
#               retAry:  成功の時、読み取ったデータ
#
#       メモ：
#---------------------------------------------------------------------------
def readSizeByType(s, typeInt, addrInt, sizeInt):
    retCode = 0
    retAry  = []
    
    addr = addrInt.to_bytes(4, 'little')
    size = sizeInt.to_bytes(4, 'little')

    # 入出力バッファクリア
    if s.in_waiting > 0:
        putLog('未処理の受信データあり: %d' % s.in_waiting)
        readStr = s.read(s.in_waiting)
        readBytes = bytes(readStr)
        putLog('未処理データ受信。')
        debugDump(readBytes)

    else:
        putLog('未受信データ無し。')

    if s.out_waiting > 0:
        putLog('送信バッファにデータあり: %d' % s.out_waiting)
        s.reset_output_buffer()
    else:
        putLog('未送信データ無し。')


    # 1. 読み取り要求コマンドに肯定応答が来るまで待つ処理
    waitCount = 0
    while True:
        # 1-1. 読み取り要求コマンド送信
        putLog('読み取り要求コマンド送信。')
        s.write(bytes([
            typeInt, 0x00, 0x00, 0x00,             # 読み込み
            0x08,    0x00, 0x00, 0x00,             # 不明 8
            size[0], size[1], size[2], size[3], # サイズ？
            addr[0], addr[1], addr[2], addr[3], # 開始アドレス
            size[0], size[1], size[2], size[3], # サイズ？
            #0x01, size[1], size[2], size[3], # サイズ？
            ]))

        # 1-2. 読み取り要求コマンド応答待ち
        respWaitCount = 0
        while True:
            if s.in_waiting > 0:
                putLog('受信データあり: %d' % s.in_waiting)
                break
            else:
                respWaitCount = respWaitCount + 1
                putLog('0.01秒待機。')
                time.sleep(0.01)
                if(respWaitCount > 40):
                    print('ERR: 読み取り要求コマンド応答タイムアウト')
                    return -1, retAry

        # 1-3. 読み取り要求コマンド応答の受信と判定
#        readStr = s.read(s.in_waiting)
        readStr = s.read(8)
        readBytes = bytes(readStr)
        putLog('読み取り要求コマンド応答受信。')
        debugDump(readBytes)

        if readBytes == bytes([0,0,0,0,size[0],size[1],size[2],size[3]]):
            putLog('肯定応答を受信。')
            break
        else:
            putLog('肯定応答でない応答を受信。')
            print('ERR: 読み取り要求コマンドに肯定応答でない応答')
            return -2, retAry

    # 2. 状態要求コマンド送信
    # 実機と異なり、ここで、いつまで待ってもデータ本体が来ない。
    # 催促のため(！？)状態要求コマンドを送信する

    # 状態要求コマンド送信
    putLog('読み取りデータ未着のため状態要求コマンド送信。')
    s.write(bytes([
        0x06, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,]))
    # 応答待ち
    waitCount = 0
    while True:
        if s.in_waiting > 0:
            putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            waitCount = waitCount + 1
            if(waitCount > 40):
                retCode = -3
                print('ERR: 読み取りデータ本体受信タイムアウト')
                return retCode, retAry
            putLog('0.01秒待機。')
            time.sleep(0.01)

    # 3. 読み取りデータ本体受信
    #   ただし、前項で状態要求を出していた場合、
    #   読み取りデータに続いて、状態応答が来るので
    #   忘れずに受信する。忘れるとシーケンスが破綻する
    readStr = s.read(sizeInt)
    readBytes = bytes(readStr)
    putLog('読み取りデータ本体受信。')
    debugDump(readBytes, addrInt)
    retAry.extend(readBytes)

    # 4. 状態要求コマンド応答受信
    readStr = s.read(12)
    readBytes = bytes(readStr)
    putLog('状態要求コマンド応答受信。')
    debugDump(readBytes)

    if readBytes == bytes([0xff,0xff,0xff,0xff,0,0,0,0]):
        putLog('ERR: 状態要求コマンド応答が不正。')
        return -4, retAry

    return retCode, retAry

#---------------------------------------------------------------------------
# 2つのbytes型を比較
#   戻り値
#        0: 一致
#       -1: 不一致。そもそも長さが違う
#       -2: 不一致。2つのbytes型の内容は異なる
#
#   メモ
#       組み込み型なのでこんなのを使わずに比較できるのでは・・・？
#---------------------------------------------------------------------------
def verifyBytes(ba1, ba2):
    if len(ba1) != len(ba2):
        print('verifyBytes(): 配列1の長さ:%dと配列2の長さ:%dが異なります。' %
            (len(ba1), len(ba2)))
        return False
    
    i = 0
    while True:
        if ba1[i] != ba2[i]:
            print('verifyBytes(): 差異検出: 配列1[%d]:%02x, 配列2[%d]:%02x' %
                (i, ba1[i], i, ba2[i]))
            return False
        
        i = i + 1
        if i == len(ba1):
            # 最後まで抜けずに一致が続いた
            return True

